"use strict";
/*
 * AST
 * Keiya Chinen @ TwoGate inc.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ts_morph_1 = require("ts-morph");
const object_path_1 = require("./object-path");
const chalk_1 = __importDefault(require("chalk"));
class AbstractSyntaxTree {
    constructor(node, pairedNode, objectPath, child) {
        this.node = node;
        this.pairedNode = pairedNode;
        this.objectPath = objectPath;
        this.child = child;
        this._valid = false;
        this.verbose = false;
    }
    get valid() {
        return this._valid;
    }
    reconstruct() {
        this.validateDescendants();
    }
    validateDescendants() {
        const node = this.node;
        if (this.verbose) {
            console.log(chalk_1.default.dim(this.objectPath.join('.')));
        }
        if (node instanceof ts_morph_1.InterfaceDeclaration || node instanceof ts_morph_1.ClassDeclaration) {
            this._valid = node.getProperties().every((m) => {
                const prop = m.getSymbol(); // IFのプロパティ
                const propName = m.getName();
                if (this.isIterableArray()) {
                    this.children = this.pairedNode.map((el, idx) => {
                        const newPath = this.objectPath.slice(0, -1).concat(idx);
                        const child = new AbstractSyntaxTree(this.node, el, newPath);
                        return child;
                    });
                    if (this.children)
                        return this.children.every((c) => c.validateDescendants());
                }
                else {
                    this.child = new AbstractSyntaxTree(m, new object_path_1.ObjectPath([propName]).traverse(this.pairedNode), this.objectPath.concat([propName]));
                    return this.child.validateDescendants();
                }
            });
            return this._valid;
        }
        else if (node instanceof ts_morph_1.PropertySignature || node instanceof ts_morph_1.PropertyDeclaration) {
            const type = node.getType();
            if (node.hasQuestionToken()) {
                if (this.pairedNode === null || this.pairedNode === undefined) {
                    this._valid = true;
                    return true;
                }
            }
            const checkResult = this.checkType(type, this.pairedNode);
            if (checkResult) {
                this._valid = true;
                return true;
            }
        }
        return this._valid;
    }
    isIterableArray() {
        const currentObjectPath = this.objectPath[this.objectPath.length - 1];
        return this.objectPath &&
            currentObjectPath instanceof Array &&
            currentObjectPath.length === 0;
    }
    checkType(type, value) {
        // console.log(
        //   "=====",type.getText(),"=====",
        //   "\nisAnonymous", type.isAnonymous(),
        //   "\nisAny",type.isAny(),
        //   "\nisArray",type.isArray(),
        //   "\nisBoolean",type.isBoolean(),
        //   "\nisBooleanLiteral",type.isBooleanLiteral(),
        //   "\nisClass",type.isClass(),
        //   "\nisClassOrInterface",type.isClassOrInterface(),
        //   "\nisEnum",type.isEnum(),
        //   "\nisEnumLiteral",type.isEnumLiteral(),
        //   "\nisInterface",type.isInterface(),
        //   "\nisIntersection",type.isIntersection(),
        //   "\nisLiteral",type.isLiteral(),
        //   "\nisNull",type.isNull(),
        //   "\nisNullable",type.isNullable(),
        //   "\nisNumber",type.isNumber(),
        //   "\nisNumberLiteral",type.isNumberLiteral(),
        //   "\nisObject",type.isObject(),
        //   "\nisString",type.isString(),
        //   "\nisStringLiteral",type.isStringLiteral(),
        //   "\nisTuple",type.isTuple(),
        //   "\nisTypeParameter",type.isTypeParameter(),
        //   "\nisUndefined",type.isUndefined(),
        //   "\nisUnion",type.isUnion(),
        //   "\nisUnionOrIntersection",type.isUnionOrIntersection(),
        //   "\nisUnknown",type.isUnknown(),
        //   "\n",value
        // )
        const unionTypes = type.getUnionTypes();
        if (type.isAny()) {
            return true;
        }
        else if (value === null) { // currently, allow null types
            return true;
        }
        else if (type.isNull() && value === null) {
            return true;
        }
        else if (type.isNumber() && typeof value === 'number') {
            return true;
        }
        else if (type.isString() && typeof value === 'string') {
            return true;
        }
        else if (type.isBoolean() && typeof value === 'boolean') {
            return true;
        }
        else if (type.isNumberLiteral() && Number(type.getText()) === value) {
            return true;
        }
        else if (type.isStringLiteral() && eval(`String(${type.getText()})`) === value) {
            return true;
        }
        else if (type.isBooleanLiteral() && (type.getText() === 'true') === value) {
            return true;
        }
        else if (type.isArray()) {
            const elType = type.getArrayElementTypeOrThrow();
            return value.every((v) => this.checkType(elType, v));
        }
        else if (unionTypes && unionTypes.length > 0) {
            return unionTypes.some((t) => this.checkType(t, value));
        }
        else if (type.isObject() && (type.isAnonymous() || type.isInterface())) {
            return type.getProperties().every((prop) => {
                const decl = prop.getDeclarations()[0];
                const propName = prop.getName();
                this.child = new AbstractSyntaxTree(decl, new object_path_1.ObjectPath([propName]).traverse(value), this.objectPath.concat([propName]));
                return this.child.validateDescendants();
            });
        }
        return false;
    }
}
exports.AbstractSyntaxTree = AbstractSyntaxTree;
//# sourceMappingURL=abstract-syntax-tree.js.map